#include <gtk/gtk.h>
#include <stdio.h>
#define _GNU_SOURCE
#include <fnmatch.h>
#include <regex.h>
#include "cat.h"

#include "search.h"
#include "utils.h"

static void find_disk(gpointer data, gpointer user_data);
static int find_entry(GNode *node, gpointer data);

static int find_entry(GNode *node, gpointer data)
{
    Item *item = ITEM(node->data);
    Search *search = (Search *)data;
    Ui *ui = search->ui;
    int rv;
    char *text[] = { NULL, NULL, NULL, NULL, NULL };
    static char *path[256];
    int i;

    if (item) {
	/* update path array */
	if (!item->is_leaf) {
	    if (path[item->level])
		g_free(path[item->level]);
	    path[item->level] = g_strdup(item->name);
	    /* let GTK update the UI */
	    while (gtk_events_pending())
		gtk_main_iteration();
	    return FALSE;
	}
	rv = compare(search->pattern, item->name, search->type, search->sens);
	if (rv == 0) {		/* found match */
	    /* file name */
	    text[0] = item->name;
	    /* disk label */
	    text[1] = search->disk->label;
	    /* path */
	    for (i = 1; i < item->level; i++) {
		char *tmp;
		tmp = text[2];
		text[2] = g_strconcat(text[2] ? text[2] : "", "/", path[i], NULL);
		g_free(tmp);
	    }
	    /* we are in root of the disk */
	    if (!text[2])
		text[2] = g_strdup("/");
	    /* file size */
	    text[3] = (item->size > 10 << 20) ?
		g_strdup_printf("%d Mb", item->size >> 20) :
		g_strdup_printf("%d", item->size);
	    /* time/date */
	    text[4] = datetime(item->time);
	    
	    gtk_clist_append(GTK_CLIST(ui->findlist), text);
	   
	    /* free what was allocated */
	    g_free(text[4]);
	    g_free(text[3]);
	    g_free(text[2]);
	}
    }
    return FALSE;
}

static void find_disk(gpointer data, gpointer user_data)
{
    Disk *disk = DISK(data);
    Search *search = (Search *)user_data;

    search->disk = disk;
    g_node_traverse(disk->tree, G_PRE_ORDER, G_TRAVERSE_ALL, 255,
	    find_entry, user_data);
}

void find_file(Search *search)
{
    Catalog *cat = search->ui->cat;

    /* empty catalog, dont bother */
    if (cat->disks == NULL)
	return;

    gtk_clist_clear(GTK_CLIST(search->ui->findlist));
    g_list_foreach(cat->disks, find_disk, search);
}

int compare(const char *pattern, const char *string, int type, int sens)
{
    regex_t *regex;
    int rv = -1;
    int err_no = 0;		/* For regerror() */
    size_t no_sub = 0;
    regmatch_t *result;
    int cflags;

    switch (type) {
	case COMPARE_EXACT:
	    if (sens)
		rv = strcmp(pattern, string);
	    else
		rv = g_strcasecmp(pattern, string);
	    break;
	case COMPARE_REGEX:
	    /* This is code inspired from the Linux Gazette Issue of July 2000,
	     * which gtktalog people copied without even understanding */
	    regex = g_new0(regex_t, 1);
	    if (sens)
		cflags = 0;
	    else
		cflags = REG_ICASE;

	    if ((err_no = regcomp(regex, pattern, cflags)) != 0) {	/* Compile the regex */
		size_t length;
		char *buffer;
		length = regerror(err_no, regex, NULL, 0);
		buffer = g_malloc(length);
		regerror(err_no, regex, buffer, length);
		fprintf(stderr, "%s\n", buffer);	/* Print the error */
		g_free(buffer);
		regfree(regex);
		return -1;
	    }
	    no_sub = regex->re_nsub + 1;	/* How many matches are there in a line? */
	    result = g_new0(regmatch_t, no_sub);
	    rv = regexec(regex, string, no_sub, result, 0);
	    regfree(regex);
	    g_free(regex);
	    g_free(result);
	    break;
	case COMPARE_GLOB:
	default:
	    if (sens)
		rv = fnmatch(pattern, string, 0);
	    else
		rv = fnmatch(pattern, string, FNM_CASEFOLD);
	    break;
    }
    return rv;
}
